#!/usr/bin/perl

use warnings;
use strict;
use File::Copy;

@ARGV != 1 and do {
    print "Usage $0 <scheduledir>\n";
    exit(-1);
};

my $scheduledir = $ARGV[0];

opendir (my $sdh, $scheduledir) or die "Failed to open dir $scheduledir ($!)\n";
my @schedules = readdir($sdh);
closedir ($sdh) or die "Failed to close dir $scheduledir ($!)\n";

chdir $scheduledir or die "Failed to change to $scheduledir ($!)\n";

foreach my $schedule (@schedules) {
    -d $schedule or next;

    $schedule =~ /^T/ and check_if_dead($schedule) and next;

    $schedule =~ /^[0-9]/ or next;

    opendir (my $sdh, $schedule) or die "Failed to open dir $schedule ($!)\n";
    my @units = readdir($sdh);
    closedir($sdh) or die "Failed to close dir $schedule ($!)\n";

    chdir $schedule or die "Failed to change to $schedule ($!)\n";

    my $num_units = 0;

    foreach my $unit (@units) {
        -d $unit or next;

        $unit =~ /^[RT0-9]/ and $num_units++;

        $unit =~ /^R/ and check_is_stall($unit) and next;
        $unit =~ /^T/ and check_if_dead($unit) and next;

        $unit =~ /^[0-9]/ or next;

        my ($schedtime, $unit_name) = split /\./, $unit, 2;
        $schedtime < time or next;

        (-e "$unit/SUSPENDED") and next;
        (-e "$unit/WANT_LGTM") and next;

        my $tempunit = "R-" . join('.', time, $$) . "-$unit";
        move $unit, $tempunit
            or next;

        ((-e "$tempunit/SUSPENDED") or (-e "$tempunit/WANT_LGTM")) and
            rename_back($unit, $tempunit) and next;

        # TODO: a units without payload should never happen; for the
        #       time being, do what is most likely to cause the least
        #       harm: ignore it.
        (-e "$tempunit/payload.bsmtp") or
            rename_back($unit, $tempunit) and next;

        system("rsmtp < \Q$tempunit/payload.bsmtp\E") == 0 or
            rename_back($unit, $tempunit) and next;

        my $tmpdir = "T-" . join('.', time, $$);
        move $tempunit, $tmpdir or
            die "Failed to move $tempunit ($!)\n";
        # Yes, we skip the rest in this casse; but if could rename once
        # but not rename any more, something really really went wrong,
        # so exit and try to avoid further damage.

        system("rm -rf \Q$tmpdir\E") == 0 or next;
        # The directory is alread renamed to a name that can be cleaned up
        # by anyone (i.e., no harm done). So, let's not stop our normal work,
        # just because someone accidently set strange permissions in a unit
        # (wrong umask or something).

        $num_units--;
    }

    chdir $scheduledir or die "Failed to change to $scheduledir ($!)\n";

    if ($num_units == 0) {
        my $tmpdir = "T-" . join('.', time, $$);
        move $schedule, $tmpdir or next;

        system("rm -rf \Q$tmpdir\E") == 0 or next;
        # The directory is alread renamed to a name that can be cleaned up
        # by anyone (i.e., no harm done). So, let's not stop our normal work,
        # just because someone accidently set strange permissions in a unit
        # (wrong umask or something).
    }
}

-d "done" and check_old_messages("done");
-d "muted" and check_muted_token("muted");

sub check_is_stall {
    my (undef, $temp_unique, $regular) = split /-/, $_[0], 3;
    my (undef, $pid) = split /\./, $temp_unique, 2;

    move $_[0], $regular unless system("ps -p $pid >/dev/null 2>&1") == 0;
    1;
}

sub check_if_dead {
    my (undef, $temp_unique) = split /-/, $_[0], 2;
    my (undef, $pid) = split /\./, $temp_unique, 2;

    system("rm -rf \Q$_[0]\E") unless system("ps -p $pid >/dev/null 2>&1") == 0;
}

sub rename_back {
    my ($sched, $tempsched) = @_;
    print "Renaming back\n";
    move $tempsched, $sched
    or die "Failed to reset $tempsched ($)!\n";
}

sub check_old_messages {
    my ($donedir) = @_;

    opendir(my $dh, $donedir) or die "Failed to open dir $donedir ($!)\n";
    my @donemessages = readdir($dh);
    closedir ($dh) or die "Failed to cloes dir $donedir ($!)\n";

    chdir $donedir or die "Failed to change to $donedir ($!)\n";

    foreach my $message (@donemessages) {

        (-f $message) or next;

        # If the message does not exist anymore, this is not a hard failure,
        # just continue.
        my @tstat = stat($message) or next;
        my $mtime = $tstat[9];

        # A message id is stored for 8 days after it was recived
        $mtime + 8*24*60*60 < time or next;

        unlink $message;
    }

    chdir $scheduledir or die "Failed to change to $scheduledir ($!)\n";
}

sub check_muted_token {
    my ($muteddir) = @_;

    opendir(my $mh, $muteddir) or die "Failed to open $muteddir ($!)\n";
    my @muteds = readdir($mh);
    closedir($mh) or die "Failed to close $muteddir ($!)\n";

    chdir $muteddir or die "Failed to change to $muteddir ($!)\n";

    foreach my $token (@muteds) {

        (-f $token) or next;

        my ($tokentimestr, undef) = split /_/, $token, 2;
        my $tokentime = (hex $tokentimestr);

        $tokentime < ( time >> 16) or next;

        unlink $token;
    }

    chdir $scheduledir or die "Failed to change to $scheduledir ($!)\n";
}
