#!/usr/bin/env python3
import email.parser
import email.policy
import email.utils
import functools
import os
import sys

import scheduledmail.general_service as general_service
import scheduledmail.util

from scheduledmail.exitcodes import EXIT_OK, EXIT_TEMP_FAIL
from scheduledmail.templates import BATCH_ANSWER_MID, PAYLOAD_MID
from scheduledmail.templates import BATCH_FAIL, BATCH_SUCCESS
from scheduledmail.templates import PURGE_HEADERS
from scheduledmail.templates import BATCH_PROLOG
from scheduledmail.unit import Unit

def failUnit(rcpt, inreply, messages, config, mid, references):
  bsmtp=BATCH_FAIL.format(
    rcpt=rcpt, inreply=inreply, mid=mid,
    references=references,
    messages=' - ' + ' - '.join(messages),
    helo = config['HELO'],
    mailfrom=config['FROM'])
  return Unit(bsmtp, name='answer')

def answer_unit(rcpt, inreply, warnings, config, mid, references):
  warn = ''
  warnstatus = ''
  if len(warnings) > 0:
    warnstatus = ' with WARNINGS'
    warn = "\n\nHowever, there were some warnings.\n - " \
           + "\n - ".join(warnings)
  bsmtp = BATCH_SUCCESS.format(
    rcpt=rcpt, inreply=inreply, warn=warn, mid=mid,
    references=references,
    warnstatus=warnstatus, helo=config['HELO'],
    mailfrom=config['FROM'])
  return Unit(bsmtp, name='answer')

def bsmtpify_mail(part, count, answeraddress, config, identifier):
  warnings = []
  name = '%d' % (count,)
  if count == 1:
    name = 'MAIN'

  # determine recipients
  tos = part.get_all('to', [])
  ccs = part.get_all('cc', [])
  bccs = part.get_all('bcc', [])
  rcptswithnames = email.utils.getaddresses(tos + ccs + bccs)
  rcpts = [mail for (_, mail) in rcptswithnames]

  if len(rcpts) == 0:
    rcpts = [answeraddress]
    warnings.append('Found mail attachment %s without recipients' % (name,))

  # purge headers
  for header in PURGE_HEADERS:
    del part[header]

  if not part.get('Message-ID'):
    part['Message-ID'] = PAYLOAD_MID.format(
      helo = config['HELO'],
      identifier = identifier,
      name = name,
    )

  msg = part.as_string()
  payload_bsmtp_lines = [
    "HELO %s" % (config["HELO"],),
    "MAIL FROM: %s" % (config["FROM"],),
  ]
  for rcpt in rcpts:
    payload_bsmtp_lines.append("RCPT TO: %s" % (rcpt,))
  payload_bsmtp_lines.append("DATA")

  for line in msg.splitlines():
    if line.startswith('.'):
      payload_bsmtp_lines.append(".%s" % (line,))
    else:
      payload_bsmtp_lines.append(line)
  payload_bsmtp_lines.append(".")
  payload_bsmtp_lines.append("QUIT")

  payload_bsmtp = "\n".join(payload_bsmtp_lines) + "\n"
  return payload_bsmtp, warnings

def process_internal(mail, mid, identifier, config, references):
  """
  Process the file for a given identifier (that may or may not be
  needed to generate answer mails) and return a list of units.
  """
  answeraddress = email.utils.parseaddr(mail.get("From"))[1]
  answermid = BATCH_ANSWER_MID.format(
     identifier=identifier, helo = config['HELO'])

  if not mail.is_multipart():
    return [failUnit(
      rcpt = answeraddress,
      refernces = references,
      inreply = mid, config = config, mid = answermid,
      references=references,
      messages = ["Nothing found to schedule. Please provide the payload as"
                  + " attached mails"],
      )]

  body = mail.get_body()
  warnings = []

  bsmtps = []
  unit_count = 0
  for part in mail.get_payload():
    if part.get_content_type() == 'message/rfc822':
      unit_count += 1
      bsmtp, part_warnings = bsmtpify_mail(
        part=part.get_payload()[0],count=unit_count,answeraddress=answeraddress,
        config=config,
        identifier=identifier
      )
      bsmtps.append(bsmtp)
      warnings.extend(part_warnings)
  bsmtp = functools.reduce(scheduledmail.util.joinBsmtp,bsmtps)

  mail.clear_content()
  del mail['cc']
  del mail['bcc']
  del mail['to']
  mail['to'] = config['SUBMIT']
  batch_prolog = BATCH_PROLOG.format(count=unit_count)
  mail.set_content(batch_prolog + body.get_content())
  mail.add_attachment(bsmtp, filename='main.bsmtp')

  final_bsmtp, final_warnings = bsmtpify_mail(part=mail, count=unit_count+1,
                                              answeraddress=answeraddress,
                                              config=config,
                                              identifier=identifier)
  main_unit = Unit(final_bsmtp)
  answer = answer_unit(rcpt=answeraddress, inreply=mid, mid=answermid,
                       references=references,
                       config=config, warnings=warnings)
  return ([main_unit, answer])


def main(argv):
   # In this service, we activate payload and answer in one act
   # so we are atomic. In particular retry does not harm, despite resubmit
   return general_service.std_service_main(
     process_internal=process_internal, argv=argv,
     exit_could_not_answer=EXIT_TEMP_FAIL, consider_rcpt=True)


if __name__ == "__main__":
  sys.exit(main(sys.argv))
