#!/usr/bin/env python3

import sys

from functools import partial

import scheduledmail.general_service as general_service
import scheduledmail.schedule

from scheduledmail.common import ScheduleError, ScheduleRetry, Success
from scheduledmail.exitcodes import EXIT_HARD_FAIL

def do_external_veto(args, config):
    answer = 'veto ' + ' '.join(args) +'\n   '
    if len(args) != 2:
        answer = answer + 'Usage: <identifier> <token>, stopping'
        return Success((False, [answer], []))
    act, _, unit = args[0].partition('/')
    if unit == '':
        unit = 'MAIN'

    try:
        veto_success = scheduledmail.schedule.external_veto(act, unit, args[1],
                                                            config=config)
    except ScheduleRetry as r:
        answer = answer + 'Needs a retry: %s' % (r,)
        return Success((False, [answer], []))
    except ScheduleError as e:
        answer = answer + 'Error during veto: %s' % (e,)
        return Success((False, [answer], []))

    answer = answer + 'Vetoed: %s/%s' % (act, unit)
    return Success((True, [answer], []), warnings=veto_success.warnings)

def do_external_unveto(args, config):
    answer = 'unveto ' + ' '.join(args) + '\n   '
    if len(args) != 2:
        answer = answer + 'Usage: <identifier> <token>, stopping'
        return Success((False, [answer], []))
    act, _, unit = args[0].partition('/')
    if unit == '':
        unit = 'MAIN'

    try:
        unveto_success = scheduledmail.schedule.external_unveto(act, unit,
                                                                args[1],
                                                                config=config)
    except ScheduleRetry as r:
        answer = answer + 'Needs a retry: %s' % (r,)
        return Success((False, [answer], []))
    except ScheduleError as e:
        answer = answer + 'Error during unveto: %s' % (e,)
        return Success((False, [answer], []))

    answer = answer + 'Unvetoed: %s/%s' % (act, unit)
    return Success((True, [answer], []), warnings=unveto_success.warnings)


def do_external_lgtm(args, config):
    answer = 'lgtm ' + ' '.join(args) + '\n   '
    if len(args) != 2:
        answer = answer + "Usage: <identifier> <token>, stopping"
        return Success((False, [answer], []))
    act, _, unit = args[0].partition('/')
    if unit == '':
        unit = 'MAIN'

    try:
        lgtm_success = scheduledmail.schedule.external_lgtm(act, unit,
                                                            args[1],
                                                            config=config)
    except ScheduleRetry as r:
        answer = answer + 'Needs a retry: %s' % (r,)
        return Success((False, [answer], []))
    except ScheduleError as e:
        answer = answer + 'Error during lgtm: %s' % (e,)
        return Success((False, [answer], []))

    answer = answer + 'LGTM: %s/%s, new stat: %s' % (act, unit,
                                                     lgtm_success.payload)
    return Success((True, [answer], []), warnings=lgtm_success.warnings)


command_operations = {'veto': do_external_veto,
                      'unveto': do_external_unveto,
                      'lgtm': do_external_lgtm}
alternatives = {}

external_process_internal = partial(
    general_service.process_commands,
    command_operations=command_operations,
    alternatives=alternatives,
    look_for_token=True)

def main(argv):
    # Here we are in deep trouble, we might have caused some side effects,
    # but cannot report about it. At the moment, we only have actions,
    # that are idempotent, but to be safe, as this seldom happens and most
    # timest indeicate, that there is a severe problem on the server, we
    # fail hard.
    # Note that we change the owner for this mail address, so we dont bounce
    # to random strangers.
    return general_service.std_service_main(
        process_internal=external_process_internal, argv=argv,
        exit_could_not_answer = EXIT_HARD_FAIL)

if __name__ == '__main__':
    sys.exit(main(sys.argv))
