#

import email.parser
import email.policy
import os
import sys
import time
import traceback

from functools import partial

import scheduledmail.parser
import scheduledmail.schedule

from scheduledmail.common import ScheduleError
from scheduledmail.config import parse_config
from scheduledmail.exitcodes import EXIT_OK, EXIT_HARD_FAIL, EXIT_TEMP_FAIL
from scheduledmail.templates import INTERNAL_ANSWER_MID, INTERNAL_ANSWER
from scheduledmail.unit import Unit


def answer_unit(*, rcpt, inreply, config, mid, answers, warnings, success,
                count_commands, count_success, references):
    warn = ''
    if len(warnings) > 0:
        warn = "\nHowever, there were some warnings.\n - " \
            + "\n - ".join(warnings)
    ans = ''
    if len(answers) > 0:
        ans = "\nYour commands produced some output.\n - " \
            + "\n - ".join(answers)
    if success:
        status = 'SUCCESS'
    else:
        status = 'FAIL'

    bsmtp = INTERNAL_ANSWER.format(
        rcpt=rcpt, inreply=inreply, references=references, warn=warn, mid=mid,
        ans=ans, helo = config['HELO'], mailfrom = config['FROM'],
        status=status, successfull=count_success, processed=count_commands)
    return Unit(bsmtp, name='answer')


def process_commands(*, command_operations, alternatives, mail, mid, identifier,
                     config, references, look_for_token=False):
    answeraddress = email.utils.parseaddr(mail.get("From"))[1]
    answermid = INTERNAL_ANSWER_MID.format(
        identifier=identifier, helo= config['HELO'])

    body = mail.get_body().get_content()
    parse_success \
        = scheduledmail.parser.parse_command_block(
            body,
            commands=list(command_operations.keys()),
            alternatives=alternatives,
            look_for_token=look_for_token,
            secret=config['secret'])

    if look_for_token:
        if parse_success is None:
            return None

        (tokens, commands) = parse_success.payload


        try:
            #TODO need to look for me before this position
            if all(
                    scheduledmail.schedule.is_token_muted(token,
                                                             answeraddress,
                                                             config=config)
                    for token in tokens):
                return None
        except Exception as e:
            # This is an internal error, so we need to temp fail at top level
            raise ScheduleError("Failed to check token for mute: %s"
                                % (e,))
    else:
        commands = parse_success.payload

    warnings = parse_success.warnings
    answers = []
    units = []
    successfull = True
    count_commands = len(commands)
    count_success = 0

    for command in commands:
        comman_success = command_operations[command[0]](command[1:],
                                                        config=config)
        (command_successful, command_answers, command_units) = \
            comman_success.payload
        command_warnings = comman_success.warnings
        answers.extend(command_answers)
        units.extend(command_units)
        warnings.extend(command_warnings)
        successfull = successfull and command_successful
        if command_successful:
            count_success += 1

    answer= answer_unit(
        rcpt=answeraddress, inreply=mid, references=references,
        config=config, mid=answermid,
        answers=answers, warnings=warnings, success=successfull,
        count_commands=count_commands, count_success=count_success)

    return units + [answer]

def default_service_process_internal(*, command_operations, alternatives):
   return partial(process_commands, command_operations=command_operations,
                  alternatives=alternatives)

def std_service_main(*, process_internal, argv,
                     exit_could_not_answer = EXIT_HARD_FAIL,
                     consider_rcpt = False):
    if len(argv) < 2:
        print("Usage: internal_service.py <config_file>...")
        return EXIT_TEMP_FAIL
    try:
        config = parse_config(argv[1:])
    except Exception as e:
        print(traceback.format_exc())
        return EXIT_TEMP_FAIL

    identifier = "%d.%d" % (int(time.time()), os.getpid())
    time.sleep(1)

    mail = email.parser.BytesParser(policy=email.policy.default)\
                       .parse(sys.stdin.buffer)
    mid = mail.get("Message-ID")

    references = ""
    if "References" in mail:
        references = mail.get("References") + " " + mid
    else:
        if "In-Reply-To" in mail:
            references = mail.get("In-Reply-To") + " " + mid
        else:
            references = mid

    doneid = mid

    if consider_rcpt:
        if "RECIPIENT" in os.environ:
            # Effective doneid is the pair of Message-Id and RECIPIENT.
            # We use an unquoted space as separator, as this is not allowed
            # in Message-Ids. Moreover, the notorious angle brackets do their
            # own means of separation.
            doneid = "%s <%s>" % (mid, os.environ["RECIPIENT"])

    try:
        if scheduledmail.schedule.check_if_done(doneid, config=config):
            # If the mail was already handled, there is no work left
            return EXIT_OK
    except Exception as e:
        # We can't be sure, that the mail is alredy done, so continue
        pass

    try:
        units = process_internal(mail=mail, mid=mid, identifier=identifier,
                                 config=config,
                                 references=references)
    except Exception as e:
        # In this case, some internal error occoured, so temp fail
        print(traceback.format_exc())
        return EXIT_TEMP_FAIL

    if units is None or len(units) == 0:
        return EXIT_OK

    try:
        status = scheduledmail.schedule.new_entry(identifier=identifier,
                                                  units=units,
                                                  config=config)
    except Exception as e:
        print("Failed, actions?!")
        print(traceback.format_exc())
        return exit_could_not_answer


    try:
        scheduledmail.schedule.message_done(doneid, status.warnings,
                                            config=config)
    except Exception as e:
        print(traceback.format_exc())
        # Here, the mail was already scheduled successfully, so we may not fail,
        # the operation has to be a success.
        pass

    return EXIT_OK
    
