#!/usr/bin/env python3
import email.utils
import sys
import time

import scheduledmail.general_service as general_service
import scheduledmail.schedule

from scheduledmail.common import Success, ScheduleError, ScheduleRetry
from scheduledmail.exitcodes import EXIT_OK, EXIT_HARD_FAIL
from scheduledmail.unit   import Unit
from scheduledmail.util import TIME_UNITS


def do_reschedule(args, config):
    answer = 'reschedule ' + " ".join(args) + '\n   '
    if len(args) < 2:
        answer = answer + 'Needs at least two arguments, stopping'
        return Success(False, [answer], [])
    act, _, unit = args[0].partition('/')
    if unit == '':
        unit = 'MAIN'

    timestring = " ".join(args[1:])
    timetuple = email.utils.parsedate_tz(timestring)
    if timetuple is None:
        answer = answer + "Failed to parse time %s, stopping" % (timestring,)
        return Success((False, [answer], []))
    time = email.utils.mktime_tz(timetuple)

    try:
        reschedule_success = \
            scheduledmail.schedule.reschedule_unit(act, unit=unit,
                                                   new_epoch=time,
                                                   config=config)
    except ScheduleRetry as r:
        answer = answer + "Needs a retry: %s" % (r,)
        return Success((False, [answer], []))
    except ScheduleError as e:
        answer = answer + "Error during reschedule: %s" % (e,)
        return Success((False, [answer], []))

    answer = answer + "Rescheduled to %d" % (time,)

    return Success((True, [answer], []), warnings=reschedule_success.warnings)

def do_reschedule_now(args, config):
    answer = 'reschedulenow ' + " ".join(args) + '\n   '
    if len(args) < 2:
        answer = answer + 'Needs at least two arguments, stopping'
        return Success(False, [answer], [])
    act, _, unit = args[0].partition('/')
    if unit == '':
        unit = 'MAIN'

    if len(args) > 3:
        answer += "Usage: reschedulenow unit count [<time unit>]"
        return Success((False, [answer], []))
    time_unit = 1
    count = 0
    if len(args) == 3:
        if args[2] in TIME_UNITS:
            time_unit = TIME_UNITS[args[2]]
        else:
            answer += "Unknown time unit %s, stopping" % (args[2],)
            return Success((False, [answer], []))
    try:
        count = int(args[1])
    except ValueError:
        answer += "Cannot parse %s as a number" % (args[1],)
        return Success((False, [answer], []))

    offset = count * time_unit
    new_time = int(time.time()) + offset

    try:
        reschedule_success = \
            scheduledmail.schedule.reschedule_unit(act, unit=unit,
                                                   new_epoch=new_time,
                                                   config=config)
    except ScheduleRetry as r:
        answer = answer + "Needs a retry: %s" % (r,)
        return Success((False, [answer], []))
    except ScheduleError as e:
        answer = answer + "Error during reschedule: %s" % (e,)
        return Success((False, [answer], []))

    answer = answer + ("Rescheduled to %d seconds from now, i.e., to %d"
                       % (offset, new_time,))

    return Success((True, [answer], []), warnings=reschedule_success.warnings)

def do_veto(args, config):
    answer = 'veto ' + ' '.join(args) + '\n   '
    if len(args) != 1:
        answer = answer + "Usage: <identifier>, stopping"
        return Success((False, [answer], []))
    act, _, unit = args[0].partition('/')
    if unit == '':
        unit = 'MAIN'

    try:
        veto_success = scheduledmail.schedule.veto(act, unit, config=config)
    except ScheduleRetry as r:
        answer = answer + "Needs a retry: %s" % (r,)
        return Success((False, [answer], []))
    except ScheduleError as e:
        answer = answer + "Error during veto: %s" % (e,)
        return Success((False, [answer], []))

    answer = answer + "Vetoed: %s/%s" % (act, unit)
    return Success((True, [answer], []), warnings=veto_success.warnings)

def do_unveto(args, config):
    answer = 'unveto ' + ' '.join(args) + '\n   '
    if len(args) != 1:
        answer = answer + "Usage: unveto <identifier>, stopping"
        return Success(False, [answer], [])
    act, _, unit = args[0].partition('/')
    if unit == '':
        unit = 'MAIN'

    try:
        unveto_success = scheduledmail.schedule.unveto(act, unit, config=config)
    except ScheduleRetry as r:
        answer = answer + "Needs a retry: %s" % (r,)
        return Success((False, [answer], []))
    except ScheduleError as e:
        answer = answer + "Error during unveto: %s" % (e,)
        return Success((False, [answer], []))

    answer = answer + "Unvetoed: %s/%s" % (act, unit)
    return Success((True, [answer], []), warnings=unveto_success.warnings)


def do_internal_mute(args, config):
    answer = 'mute ' + ' '.join(args) + "\n   "
    if len(args) != 1:
        answer = answer + "Usage: <token>, stopping"
        return Success((False, [answer], []))

    try:
        mute_success = scheduledmail.schedule.mute_token(args[0], config=config)
    except ScheduleRetry as r:
        answer = answer + "Needs a retry: %s" % (r,)
        return Success((False, [answer], []))
    except ScheduleError as e:
        answer = answer + "Error during mute: %s" % (e,)
        return Success((False, [answer], []))

    answer = answer + "Muted: %s" % (token,)
    return Success((True, [answer], []), warnings=mute_success.warnings)


def do_internal_remove(args, config):
    answer = 'remove ' + ' '.join(args) + '\n   '
    if len(args) != 1:
        answer = answer + "Usage: remove <identifier>, stopping"
        return Success((False, [answer], []))
    act, _, unit = args[0].partition('/')
    if unit == '':
        unit = 'MAIN'

    try:
        remove_success = scheduledmail.schedule.remove(act, unit, config=config)
    except ScheduleRetry as r:
        answer = answer + "Needs a retry: %s" % (r,)
        return Success((False, [answer], []))
    except ScheduleError as e:
        answer = answer + "Error during remove: %s" % (e,)
        return Success((False, [answer], []))

    answer = answer + "Removed: %s/%s" % (act, unit)
    return Success((True, [answer], []), warnings=remove_success.warnings)


def do_internal_lgtm(args, config):
    answer = 'lgtm ' + ' '.join(args) + '\n   '
    if len(args) < 1 or len(args) > 2:
        answer = answer + "Usage: lgtm <identifier> [count], stopping"
        return Success((False, [answer], []))
    elif len(args) == 1:
        count = '1'
    else:
        count = args[1]
    act, _, unit = args[0].partition('/')
    if unit == '':
        unit = 'MAIN'

    try:
        lgtm_success = scheduledmail.schedule.lgtm(act, unit, count=count,
                                                   config=config)
    except ScheduleRetry as r:
        answer = answer + "Needs a retry: %s" %(r,)
        return Success((False, [answer], []))
    except ScheduleError as e:
        answer = answer + "Error during lgtm: %s" % (e,)
        return Success((False, [answer], []))

    answer = answer + "LGTM: %s/%s, new state: %s" % (act, unit,
                                                      lgtm_success.payload)
    return Success((True, [answer], []), warnings=lgtm_success.warnings)


command_operations = {'reschedule': do_reschedule,
                      'reschedulenow': do_reschedule_now,
                      'veto': do_veto,
                      'unveto': do_unveto,
                      'mute': do_internal_mute,
                      'remove': do_internal_remove,
                      'lgtm': do_internal_lgtm}
alternatives = { 'suspend': 'veto',
                 'unsuspend': 'unveto',
                 'now': 'reschedulenow',
                 }


def main(argv):
    internal_process = general_service.default_service_process_internal(
        command_operations=command_operations, alternatives=alternatives)
    # Here we are in deep trouble, we might have caused some side effects,
    # but cannot report about it. At the moment, we only have actions,
    # that are idempotent, but to be safe, as this seldom happens and most
    # timest indeicate, that there is a severe problem on the server, we
    # fail hard.
    # Note that the sender is necessarily one with access to the local
    # domains, hence we are not bouncing to random strangers.
    return general_service.std_service_main(
        process_internal=internal_process, argv=argv,
        exit_could_not_answer = EXIT_HARD_FAIL)

if __name__ == "__main__":
  sys.exit(main(sys.argv))
