#

import os
import pathlib
import subprocess
import time

import scheduledmail.config as mailconfig
from scheduledmail.common import Success, ScheduleError, ScheduleRetry
from scheduledmail.scheduleutil import current_simple_unit, scheduledir_mgr
from scheduledmail.scheduleutil import current_unit, update_want_lgtm
from scheduledmail.util import simpleQuote

def runqueueat(epochtime, scheduledir, config=None):
  """
  Schedule a run of the queue-handler for the given directory
  at the given time.

  @rtype: Success(None)
  """
  if config is None:
    config = {}
  # Avoid trouble with minute-precision for some at(1)s
  attime = time.strftime("%Y%m%d%H%M", time.localtime(epochtime + 61))
  at = mailconfig.AT
  if "at" in config:
    at = config["at"]
  cmd = [at, "-t", attime]
  queuerunner = mailconfig.QUEUERUNNER
  if "queuerunner" in config:
    queuerunner = config["queuerunner"]
  if isinstance(queuerunner, str):
    queuerunner = queuerunner.encode()
  if isinstance(scheduledir, str):
    scheduledir = scheduledir.encode()
  atinput = b"%s '%s'" % (queuerunner, scheduledir.replace(b"'", b"'\\''"))

  process = subprocess.Popen(cmd, stdin=subprocess.PIPE)
  process.stdin.write(atinput)
  process.stdin.close()
  retcode = process.wait()
  if retcode != 0:
    return Success(None,
                   warnings = ["%s with input %s returned %d" %
                               (cmd, atinput, retcode)],
                   dirruns = [epochtime])
  return Success(None)

def new_entry(identifier, units=None, config=None):
  """
  Generate a new entry for to schedule mail. Also schedule
  the call to at(1).

  @raises: ScheduleError
  @rtype: Success(String)
  @return: the id of the sceduled mail
  """
  if units is None or not units:
    return Success(identifier)
  if config is None:
    config = {}

  scheduledir = mailconfig.SCHEDULEDIR
  if "scheduledir" in config:
    scheduledir = config["scheduledir"]
  if not os.path.isdir(scheduledir):
    raise ScheduleError("Schedule directory '%s' does not exist"
                        % (scheduledir,))
  
  tmpdirname = os.path.join(scheduledir, "T-%s" %(identifier,))
  finaldirname = os.path.join(scheduledir, "%s" %(identifier,))

  try:
    os.mkdir(tmpdirname)
    for unit in units:
      create_unit(tmpdirname, unit)
    os.sync()
  except Exception as e:
    raise ScheduleError("Error creating schedule on disk: %s" % (e,))

  try:
    os.rename(tmpdirname, finaldirname)
    os.sync()
  except OSError as e:
    raise ScheduleError("Error activating '%s' as '%s': %s"
                        % (tmpdirname, finaldirname, e))

  # for each epoch time of a unit runqueueat
  times = list(set([unit.epoch for unit in units]))
  return Success(identifier).after(
    Success.forM_(times,
                  lambda t: runqueueat(t, scheduledir, config)))


def create_unit(directory, unit):
  """
  Create a unit in the given directory. Note that new units may only be
  created in temporary act directories.
  """
  finaldirname = os.path.join(directory, "%s.%s" % (unit.epoch, unit.name))

  try:
    os.mkdir(finaldirname)
    with open(os.path.join(finaldirname, "payload.bsmtp"), "w") as f:
      f.write(unit.payload)
    for file_name, content in unit.extra_files.items():
      with open(os.path.join(finaldirname, file_name), "w") as f:
        f.write(content)
    os.sync()
  except Exception as e:
    raise ScheduleError("Error creating unit on disk: %s" %(e,))

def message_done(message, warnings, config=None):
  """
  Create the message done file.
  """
  if config is None:
    config = {}

  scheduledir = mailconfig.SCHEDULEDIR
  if "scheduledir" in config:
    scheduledir = config["scheduledir"]
  if not os.path.isdir(scheduledir):
    raise ScheduleError("Schedule directory '%s' does not exist"
                        % (scheduledir,))

  donedir = os.path.join(scheduledir, "done")
  if not os.path.isdir(donedir):
    try:
      os.mkdir(donedir)
    except Exception as e:
      raise ScheduleError("Failed to create done dir '%s': %s"
                          % (donedir, e))
  quoted_mid = simpleQuote(message)

  try:
    with open(os.path.join(donedir, quoted_mid), "w") as  f:
      for warning in warnings:
        f.write("- %s\n" % (warning,))
  except Exception as e:
    raise ScheduleError("Failed to create done for message '%s': %s"
                        % (message, e))

def check_if_done(message, config=None):
  """
  Check if message was already done.
  """
  if config is None:
    config = {}

  scheduledir = mailconfig.SCHEDULEDIR
  if "scheduledir" in config:
    scheduledir = config["scheduledir"]
  if not os.path.isdir(scheduledir):
    raise ScheduleError("Schedule directory '%s' does not exist"
                        % (scheduledir,))

  donedir = os.path.join(scheduledir, "done")
  if not os.path.isdir(donedir):
    return False
  quoted_mid = simpleQuote(message)

  try:
    return os.path.isfile(os.path.join(donedir, quoted_mid))
  except Exception as e:
    raise ScheduleError("Could not check done '%s': %s"
                        % (quoted_mid, e))

def reschedule_unit(act, unit = 'MAIN', new_epoch=0, config=None):
  """
  Reschedules a pending unit to a new time.
  """
  if config is None:
    config = {}

  scheduledir = mailconfig.SCHEDULEDIR
  if "scheduledir" in config:
    scheduledir = config["scheduledir"]
  actdir = os.path.join(scheduledir, act)
  new_unit = os.path.join(actdir, "%s.%s" % (new_epoch, unit))

  with current_simple_unit(act, unit, config=config) as old_unit:

    try:
      old_unit.rename(new_unit)
      os.sync()
    except Exception as e:
      raise ScheduleRetry("Could not reschedule '%s', needs retry: %s"
                          % (unit, e))


  return runqueueat(new_epoch, scheduledir, config=config)

def veto(act, unit='MAIN', config=None):

  with current_simple_unit(act, unit, config=config) as current_unit:

    try:
      (current_unit / 'SUSPENDED').touch()
      os.sync()
    except Exception as e:
      raise ScheduleRetry("Could not veto '%s', needs retry: %s"
                          % (unit, e))

  return Success(None)


def unveto(act, unit='MAIN', config=None):

  with current_simple_unit(act, unit, config=config) as current_unit:

    try:
      (current_unit / 'SUSPENDED').unlink()
      os.sync()
    except FileNotFoundError as fnf:
      # There was a moment, when the file doesn't exist, so we where successfull
      pass
    except Exception as e:
      raise ScheduleRetry("Could not veto '%s', needs retry: %s"
                          % (unit, e))

  return Success(None)

def remove(act, unit='MAIN', config=None):

  with scheduledir_mgr(config) as scheduledir:
    actdir = pathlib.Path(os.path.join(scheduledir, act))

  tmp_identifier = "%d.%d" % (int(time.time()), os.getpid())

  with current_simple_unit(act, unit, config=config) as current_unit:

    try:
      tmp_unit = actdir / ('T-%s-%s' % (tmp_identifier, current_unit.name))
      current_unit.rename(tmp_unit)
      os.sync()
    except Exception as e:
      raise ScheduleRetry("Could not remove '%s', needs retry: %s"
                          % (unit, e))

  return Success(None)

def is_token_muted(token, answer, config):

  with scheduledir_mgr(config) as scheduledir:
    mutedir = pathlib.Path(scheduledir) / 'muted'

    try:
      if not mutedir.is_dir():
        mutedir.mkdir()
    except Exception as e:
      raise ScheduleError("Error with mutedir: %s" % (e,))

    quoted_token = simpleQuote(token)
    tokenfile = mutedir / quoted_token

    try:
      if not tokenfile.is_file():
        return False

      with tokenfile.open('r') as tf:
        line = tf.readline()
        if line == answer:
          return False
        else:
          return True
    except Exception as e:
      raise ScheduleError("Error during lookup of token: %s"
                          % (e,))


def mute_token(token, config, answer=None):
 with scheduledir_mgr(config) as scheduledir:
    mutedir = pathlib.Path(scheduledir) / 'muted'

    try:
      if not mutedir.is_dir():
        mutedir.mkdir()
    except Exception as e:
      raise ScheduleError("Error with mutedir: %s" % (e,))

    quoted_token = simpleQuote(token)
    tokenfile = mutedir / quoted_token

    try:
      # TODO currently we only mute unconditionally
      # Maybe we later want to safe the answer address
      tokenfile.touch()
    except Exception as e:
      raise ScheduleError


def is_token_valid_unit(current_unit, token, role, config):
  quoted_token = simpleQuote(token)
  tokenfile = current_unit / ('token-%s' % (quoted_token,))

  try:
    if not tokenfile.is_file():
      return False

    with tokenfile.open('r') as tf:
      line = tf.readline()
      # TODO this only allows single purpose tokens
      return role == line.strip()
  except Exception as e:
    raise ScheduleError("Could not test validity of token: %s" % (e,))


def invalidate_token(current_unit, token, config):
  quoted_token = simpleQuote(token)
  tokenfile = current_unit / ('token-%s' % (quoted_token,))

  try:
    tokenfile.unlink()
  except Exception as e:
    raise ScheduleError('Could not invalidate token: %s' % (e,))


def external_veto(act, unit, token, config):
  #TODO this should use curren_unit for more
  # safety
  with current_simple_unit(act, unit, config=config) as current_unit:
    try:
      if not is_token_valid_unit(current_unit, token, 'veto',
                                 config=config):
        raise ScheduleError('Token not valid for task')
    except Exception as e:
      raise ScheduleError('Could not test validity of token: %s' % (e,))

    try:
      (current_unit / 'SUSPENDED').touch()
      os.sync()
    except Exception as e:
      raise ScheduleRetry("Could not veto '%s', needs retry: %s"
                          % (unit, e))

    try:
      # TODO new token?
      invalidate_token(current_unit, token, config=config)
      # TODO this mutes unconditionally...
      mute_token(token, config=config)
    except Exception as e:
      return Success(None, warnings=[e])

    return Success(None)


def external_unveto(act, unit, token, config):
  #TODO this should use curren_unit for more
  # safety
  with current_simple_unit(act, unit, config=config) as current_unit:
    try:
      if not is_token_valid_unit(current_unit, token, 'unveto',
                                 config=config):
        raise ScheduleError('Token not valid for task')
    except Exception as e:
      raise ScheduleError('Could not test validity of token: %s' % (e,))

    try:
      (current_unit / 'SUSPENDED').unlink()
      os.sync()
    except Exception as e:
      raise ScheduleRetry("Could not unveto '%s', needs retry: %s"
                          % (unit, e))

    try:
      # TODO new token?
      invalidate_token(current_unit, token, config=config)
      # TODO this mutes unconditionally...
      mute_token(token, config=config)
    except Exception as e:
      return Success(None, warnings=[e])

    return Success(None)


def lgtm(act, unit, count, config):
  with current_unit(act, unit, config=config) as current:
      return update_want_lgtm(current, count, config)


def external_lgtm(act, unit, token, config):
  with current_unit(act, unit, config=config) as current:
    try:
      if not is_token_valid_unit(current, token, 'lgtm',
                                 config=config):
        raise ScheduleError('Token not valid for task')
    except Exception as e:
      raise ScheduleError('Could not test validity of token: %s' % (e,))

    lgtm_success = update_want_lgtm(current, lgtms="1", config=config)

    try:
      # TODO new token?
      invalidate_token(current, token, config=config)
      # TODO this mutes unconditionally...
      mute_token(token, config=config)
    except Exception as e:
      return Success(None, warnings=[e]).than(lgtm_success)

    return lgtm_success
