#

import os
import time

from contextlib import contextmanager
from pathlib import Path

import scheduledmail.config as mailconfig

from scheduledmail.common import ScheduleError, ScheduleRetry, Success
from scheduledmail.lgtm import do_lgtm, format_lgtm_status

@contextmanager
def scheduledir_mgr(config):
  if "scheduledir" in config:
    scheduledir = config["scheduledir"]
  else:
    raise ScheduleError("Missing scheduledir in config: %s"
                        % (config,))
  if not os.path.isdir(scheduledir):
      raise ScheduleError("Schedule directory '%s' does not exist"
                          % (scheduledir,))
  yield scheduledir

@contextmanager
def current_simple_unit(act, unit, config):
 with scheduledir_mgr(config) as scheduledir:

   actdir = os.path.join(scheduledir, act)

   if not os.path.isdir(actdir):
     raise ScheduleError("Act '%s' does not exist"
                         % (act,))

   try:
     current_units = os.listdir(actdir)
   except Exception as e:
     raise ScheduleError("Error reading act '%s': %s"
                         % (act, e))
   matching_units = \
     list(filter(lambda u: u.endswith(".%s" % (unit,)), current_units))
   if len(matching_units) == 0:
     raise ScheduleError("Unit '%s' does not exist at all" % (unit,))
   if len(matching_units) > 1:
     raise ScheduleError("Unit '%s' exist multiple times" % (unit,))
   current_unit = matching_units[0]
   if current_unit.startswith('T-'):
     raise ScheduleError("Unit '%s' does not exist at all" % (unit,))
   if current_unit.startswith('R-'):
     raise ScheduleRetry("Unit '%s' unter run, needs retry" % (unit,))

   yield (Path(actdir) / current_unit)


@contextmanager
def current_unit(act, unit, config):

  with scheduledir_mgr(config) as scheduledir:
    actdir = Path(os.path.join(scheduledir, act))

  tmp_identifier = "%d.%d" % (int(time.time()), os.getpid())

  with current_simple_unit(act, unit, config) as current_unit:

    try:
      tmp_unit = actdir / ('R-%s-%s' % (tmp_identifier, current_unit.name))
      current_unit.rename(tmp_unit)
    except Exception as e:
      raise ScheduleRetry("Failed to set unit %s under run, needs retry: %s"
                          % (unit, e))

    yield tmp_unit

    try:
      tmp_unit.rename(current_unit)
    except Exception as e:
      # This is not an error, since mail-schedule-do will take care of
      # renaming the run dir back after we died
      pass


def update_want_lgtm(current_unit, lgtms, config):
  """Given a unit currently under lock and a count of lgtms that should
  be applied, update the WANT_LGTM file.
  """

  want_lgtm_file = current_unit / 'WANT_LGTM'
  tmp_file = current_unit / 'T-WANT_LGTM'
  warnings = []

  if want_lgtm_file.is_file():
    try:
      want_lgtm = want_lgtm_file.read_text()
    except Exception as e:
      # If it is not possible to read the WANT_LGTM file pretend that
      # it's content is the empty string. The empty string is treated
      # as one LGTM required, what seems to be a reasonable default as
      # an existing (but not readable) WANT_LGTM file blocks the unit.
      want_lgtm = ""
  else:
    want_lgtm = None

  new_want_lgtm = do_lgtm(want_lgtm, lgtms)

  if new_want_lgtm is None:
    if not want_lgtm_file.is_file():
      return Success(format_lgtm_status(None), warnings=warnings)
    else:
      try:
        want_lgtm_file.unlink()
      except Exception as e:
        raise ScheduleError('Failed to remove WANT_LGTM file: %s' % (e,))

      return Success(format_lgtm_status(None), warnings=warnings)

  try:
    tmp_file.write_text(new_want_lgtm)
    os.sync()
    tmp_file.rename(want_lgtm_file)
    os.sync()
  except Exception as e:
    raise ScheduleError('Failed to update WANT_LGTM file: %s' % (e,))

  return Success(format_lgtm_status(new_want_lgtm), warnings=warnings)
