#!/usr/bin/env python3

import calendar
import os
import sys
import time

import scheduledmail.general_service as general_service

from scheduledmail.exitcodes import EXIT_TEMP_FAIL
from scheduledmail.submitutil import answerUnit
from scheduledmail.templates import SUBMISSION_ANSWER_MID
from scheduledmail.unit import Unit

def unquote_and_split(quotedmails):
  parts = quotedmails.replace(':', '.').split('%')
  res = []
  if parts:
    res.append(parts.pop(0))
  while parts:
    v = parts.pop(0)
    try:
      res.append(chr((int(v[:2], 16) % 128)))
    except Exception:
      # the quoting was done incorrectly; so, for lack of
      # alternatives, assume the % was meant literally. At
      # least, we should not crash.
      res.append(v)
    res.append(v[2:])
  return ''.join(res).split('\x00')

def split_time_rcpt(ext):
  timercpts = ext.split('-', 1)
  if len(timercpts) < 2:
    return timercpts[0], []
  time = timercpts[0]
  rcpts = timercpts[1]
  return timercpts[0], unquote_and_split(timercpts[1])

def process_internal(mail, mid, identifier, config, references):
  answermid = SUBMISSION_ANSWER_MID.format(
    identifier=identifier, helo = config["HELO"])

  remindertime = 0
  rcpts = []
  if "EXT4" in os.environ:
    ext, rcpts = split_time_rcpt(os.environ["EXT4"])
    formatstring = None
    if len(ext) == 8:
      # Add the current year, if no year is given; otherwise we'll
      # end up in python's default year 1900
      ext = "%s%s" % (time.strftime("%Y"), ext)
    if len(ext) == 10:
      formatstring = "%y%m%d%H%M"
    if len(ext) == 12:
      formatstring = "%Y%m%d%H%M"
    if formatstring:
      timestruct = time.strptime(ext, formatstring)
      remindertime = calendar.timegm(timestruct)

  extra_files = {}
  desc = ""
  if "Subject" in mail:
    desc += "Subject: %s\n" % (mail.get("Subject"),)
  if "From" in mail:
    desc += "From: %s\n" % (mail.get("From"),)
  if "To" in mail:
    desc += "To: %s\n" % (mail.get("To"),)
  if desc:
    # The description might contain fancy characters. For our description file,
    # we replace everything that is not ascii (which is the encoding we use to
    # get bytes out of the string when writing the file) by a simple question
    # mark. Should be good enough for a human-readable summary.
    newdesc = ""
    for c in desc:
      if ord(c) < 128:
        newdesc += c
      else:
        newdesc += '?'
    extra_files["DESCRIPTION"] = newdesc

  reminder_for = config["SIMPLE_TO"]
  if "SIMPLE_X_REMINDER_FOR" in config:
    reminder_for = config["SIMPLE_X_REMINDER_FOR"]

  payload_bsmtp_lines = [
    "HELO %s" % (config["HELO"],),
    "MAIL FROM: %s" % (config["SIMPLE_FROM"],),
    "RCPT TO: %s" % (config["SIMPLE_TO"],),
  ] + ["RCPT TO: %s" % (rcpt,) for rcpt in rcpts] + [
    "DATA",
    "X-reminder-for: %s" % (reminder_for,),
  ]
  for line in mail.as_string().splitlines():
    if line.startswith('.'):
      payload_bsmtp_lines.append(".%s" % (line,))
    else:
      payload_bsmtp_lines.append(line)
  payload_bsmtp_lines.append(".")
  payload_bsmtp_lines.append("QUIT")
  payload_bsmtp = "\n".join(payload_bsmtp_lines) + "\n"
  main_unit = Unit(payload_bsmtp, epoch=remindertime, extra_files=extra_files)

  # Generate an answer to the log email address, so that the internal identifier
  # is also available offline in the synchronized mail box.
  answer_unit = answerUnit(
    rcpt=config["SIMPLE_LOG"],
    cc=[],
    inreply=mid,
    references=references,
    warnings=[],
    config=config,
    mid = answermid,
    identifier = identifier,
    time = remindertime,
    extrafiles = extra_files,
  )

  return [main_unit, answer_unit]

def main(argv):
  # In this service, we activate payload and answer in one act
  # so we are atomic. In particular retry does not harm, despite resubmit
  return general_service.std_service_main(
    process_internal=process_internal, argv=argv,
    exit_could_not_answer=EXIT_TEMP_FAIL, consider_rcpt=True)

if __name__ == "__main__":
  sys.exit(main(sys.argv))
