#!/usr/bin/env python3
import email.parser
import email.policy
import email.utils
import os
import sys
import time

import scheduledmail.general_service as general_service
import scheduledmail.mailtoken
import scheduledmail.parser
import scheduledmail.schedule
import scheduledmail.util

from scheduledmail.common import Success
from scheduledmail.exitcodes import EXIT_OK, EXIT_TEMP_FAIL
from scheduledmail.lgtm import format_lgtm_status
from scheduledmail.unit import Unit
from scheduledmail.util import simpleQuote, TIME_UNITS
from scheduledmail.submitutil import answerUnit, cc_header, cc_rcpt
from scheduledmail.templates import SUBMIT_FAIL, SCHEDULED_MAIL_SENT
from scheduledmail.templates import SUBMISSION_ANSWER_MID, SCHEDULED_MAIL_SENT_MID, SCHEDULED_MAIL_PART_SENT_MID
from scheduledmail.templates import TOKEN_DESC, TOKEN_MAIL, TOKEN_MAIL_MID, PART_DESC_ANNOUNCE
from scheduledmail.templates import ANNOUNCE, ANNOUNCE_MID
from scheduledmail.templates import PAYLOAD_MID
from scheduledmail.templates import FINAL_CMD_MAIL_MID, FINAL_CMD_MAIL
from scheduledmail.templates import PURGE_HEADERS

TOKEN_EXTENDED_VALIDITY = 7 * 24 * 60 * 60

CMD_ALTERNATIVES = { 'veto': 'suspend',
                     'desc': 'description',
                     '!': 'description',
                     'warn' : 'announce',
                     'warnmsg' : 'announcemsg',
                     '!!' : 'announcemsg',
                     'want_lgtm' : 'wantlgtm',
}

def get_time(commands):
  """
  From a list of commands, determine the time in seconds
  since the epoch at which to schedule the mail, of None,
  if this failed. Do so in the Success writer Monad.
  """
  time = None
  warnings = []
  for cmd in commands:
    if cmd[0] != "schedule":
      continue
    if len(cmd) < 2:
      warnings.append("schedule command expects an argument")
      contine
    timestring = " ".join(cmd[1:])
    timetuple = email.utils.parsedate_tz(timestring)
    if timetuple is None:
      warnings.append("schedule: failed to parse time '%s'" % (timestring,))
      continue
    time = email.utils.mktime_tz(timetuple)
  return Success(time, warnings=warnings)

def not_before_now(t):
  """
  Given a time, in seconds since the epoch, compare to the current time and
  return
  - the later of the two times, and
  - a list of warnings about that time being in the past.
  """
  now = int(time.time())
  if t >= now:
    return t, []
  offset = now - t
  msg = ("Specified time is %.1f days in the past."
         % (float(offset) / float(24 * 60 * 60),))
  return now, [msg]

def get_me(commands):
  me = None
  warnings = []
  for cmd in commands:
    if cmd[0] != "me":
      continue
    if len(cmd) != 2:
      warnings.append("me needs precisely one argument, found %s" % (cmd,))
      continue
    me = cmd[1]
  return Success(me, warnings=warnings)

def get_cc(commands):
  cc = []
  for cmd in commands:
    if cmd[0] == "cc":
      cc.extend(cmd[1:])
  return cc

def failUnit(*, rcpt, cc, inreply, messages, config, mid, references):
  bsmtp=SUBMIT_FAIL.format(
    rcpt=rcpt, inreply=inreply, mid=mid, references=references,
    messages=" - " + " - ".join(messages),
    helo = config["HELO"],
    cc_hdr=cc_header(cc), additional_rcpt=cc_rcpt(cc),
    mailfrom = config["FROM"])
  return Unit(bsmtp, name="answer")

def extra_files(cmds):
  result = {}
  cc = get_cc(cmds)
  if len(cc) > 0:
    result["CC"] = "\n".join(cc) + "\n"
  descs = [cmd[1:] for cmd in cmds if cmd[0] == "description"]
  if len(descs) > 0:
    result["DESCRIPTION"] = "".join([" ".join(desc) + "\n" for desc in descs])
    result["DESCRIPTION"] = result["DESCRIPTION"] \
      .encode('ascii', 'replace').decode('ascii')
  suspends = [cmd[1:] for cmd in cmds if cmd[0] == "suspend"]
  if len(suspends) > 0:
    result["SUSPENDED"] = "".join([" ".join(suspend) + "\n"
                                   for suspend in suspends])
    result["SUSPENDED"] = result["SUSPENDED"] \
      .encode('ascii', 'replace').decode('ascii')
  for cmd in cmds:
    if cmd[0] == "wantlgtm":
        result["WANT_LGTM"] = " ".join(cmd[1:])
  return result

def generate_tokens(*, cmds, identifier, count, secret, validuntil,
                    ismain=True):
  """
  Generate a fresh set of tokens, as dictated by the command.

  Return a pair of
  - the new count
  - a list of (mail, use, token) tripples
  """
  result = []
  for cmd in cmds:
    if ( (cmd[0] == "token" or (cmd[0] == "tokenmain" and ismain))
         and len(cmd) > 2 and cmd[1] in TOKEN_DESC):
      for mail in cmd[2:]:
        count += 1
        token = scheduledmail.mailtoken.generate_token(
          secret=secret, identifier=identifier, count=count,
          validuntil=validuntil)
        result.append((mail, cmd[1], token))
  return (count, result)

def act_desc(*, extra_files, identifier):
  if "DESCRIPTION" in extra_files:
    lines = [ "",
              "Scheduled mail %s was described as follows." % (identifier,),
              "" ]
    for line in extra_files["DESCRIPTION"].splitlines():
      # If changing the quote prefix, remember to keep it bsmtp quoted.
      lines.append(" | %s" % (line,))
    lines.append("")
    lines.append("")
    return "\n".join(lines)
  else:
    return ""

def generateAnnounce(*, epoch, identifier, cmds, config, answeraddress,
                     references):
  warnings = []
  offset = None
  origcmd = None
  inreply=SUBMISSION_ANSWER_MID.format(identifier=identifier,
                                       helo=config["HELO"])
  for cmd in cmds:
    if cmd[0] == "announce":
      if len(cmd) < 2:
        warnings.append("Ignoring 'announce' without arguments")
        continue
      if len(cmd) > 3:
        warnings.append("Ignoring %s as it has too many arguments" % (cmd,))
        continue
      if len(cmd) == 2:
        unit = 1
      else:
        if cmd[2] in TIME_UNITS:
          unit = TIME_UNITS[cmd[2]]
        else:
          warnings.append("Ignoring %s, unkown time units" % (cmd,))
          continue
      try:
        offset = int(cmd[1]) * unit
        origcmd = cmd
      except ValueError:
        warnings.append("Ignoring %s, cannot parse number" % (cmd,))
  if offset is None:
    return Success([], warnings=warnings)
  extrafiles = extra_files(cmds)
  actdesc = act_desc(extra_files=extrafiles, identifier=identifier)
  mydesc = "Announce mail for %s" % (identifier)
  if "DESCRIPTION" in extrafiles:
    mydesc += " which was described as follows.\n\n" \
              + extrafiles["DESCRIPTION"]
  else:
    mydesc += "\n";
  extrafiles["DESCRIPTION"] = mydesc
  suspmsg =""
  if "SUSPENDED" in extrafiles:
    suspmsg = "\n".join(["",
                         "At submission, the %s was suspended." % (identifier,)]
                        + [ " | " + line
                            for line in extrafiles["SUSPENDED"].splitlines()]
                        + [""])
    del extrafiles["SUSPENDED"]
  if "WANT_LGTM" in extrafiles:
    suspmsg += ("\nAt submission %s wanted LGTMs: %s.\n"
                % (identifier,format_lgtm_status(extrafiles["WANT_LGTM"])))
    del extrafiles["WANT_LGTM"]
  announcemsgs = []
  for cmd in cmds:
    if cmd[0] == "announcemsg":
      announcemsgs.append(" ".join(cmd[1:]))
  msg = ""
  if len(announcemsgs) > 0:
    msg = "\nYou also asked me to tell you the following.\n" \
          + "\n".join([" | " + line for line in announcemsgs]) \
          + "\n"
  cc = get_cc(cmds)
  bsmtp = ANNOUNCE.format(
    epoch=epoch,
    helo = config["HELO"],
    mailfrom = config["FROM"],
    rcpt = answeraddress,
    additional_rcpt=cc_rcpt(cc),
    cc_hdr=cc_header(cc),
    mid = ANNOUNCE_MID.format(identifier=identifier, helo=config["HELO"]),
    inreply=inreply,
    references = references + " " + inreply,
    identifier=identifier,
    origcmd=origcmd,
    msg=msg,
    actdesc=actdesc,
    suspmsg=suspmsg)
  return Success([Unit(bsmtp, epoch=epoch-offset, name="announce",
                       extra_files=extrafiles)],
                 warnings=warnings)

def generate_final_mail(*, identifier, name, final_cmds, answeraddress, config,
                        references):
  mid = FINAL_CMD_MAIL_MID.format(
    identifier=identifier, name=name, helo=config["HELO"])
  submission_mid = SUBMISSION_ANSWER_MID.format(
    identifier=identifier, name=name, helo=config["HELO"])
  inreply = SCHEDULED_MAIL_PART_SENT_MID.format(
    identifier=identifier, name=name, helo=config["HELO"])

  cmdblockquoted = "\n".join(
    [ line if not line.startswith(".") else "." + line
      for line in [
          " ".join(words) for words in final_cmds
  ]])
  return FINAL_CMD_MAIL.format(
    helo = config["HELO"],
    user = answeraddress,
    internalcontrol = config["INTERNAL_CONTROL"],
    mid = mid,
    inreply = inreply,
    references = references + " " + submission_mid + " " + inreply,
    identifier = "%s/%s" % (identifier, name),
    cmdblockquoted = cmdblockquoted,
  )

def process_bsmtp(*, identifier, config, bsmtp, answeraddress, answermid,
                  mail, mid, references):
  body = mail.get_body().get_content()
  commands_success = scheduledmail.parser.parse_command_block(
    body, commands=["schedule", "me", "cc", "suspend", "description", "token",
                    "announce", "announcemsg", "wantlgtm", "finally"],
    alternatives=CMD_ALTERNATIVES)

  commands = commands_success.payload
  warnings = commands_success.warnings
  time_success = get_time(commands)
  time = time_success.payload
  warnings.extend(time_success.warnings)
  me_success = get_me(commands)
  warnings.extend(me_success.warnings)
  if me_success.payload:
    answeraddress = me_success.payload
  cc = get_cc(commands)

  if time is None:
    return [failUnit(
      rcpt = answeraddress, cc=cc,
      inreply = mid, references=references, config = config, mid = answermid,
      messages = ["Could not determine the time to schedule the request"]
                  + warnings,
    )]

  announceunits_succ = generateAnnounce(epoch=time, identifier=identifier,
                                        references=references,
                                        cmds=commands, config=config,
                                        answeraddress=answeraddress)
  warnings.extend(announceunits_succ.warnings)
  announceunits = announceunits_succ.payload


  cmd_extra_files = extra_files(commands)

  referencetime, timewarnings = not_before_now(time)
  warnings.extend(timewarnings)

  answer = answerUnit(
    rcpt=answeraddress, cc=cc, inreply=mid, references=references,
    config=config, mid=answermid,
    identifier=identifier, time=time,
    warnings=warnings,
    names = ["MAIN"] + [au.name for au in announceunits],
    extrafiles = cmd_extra_files,
  )

  allsentbsmtp = SCHEDULED_MAIL_SENT.format(
    helo = config["HELO"], mailfrom = config["FROM"],
    mid = SCHEDULED_MAIL_SENT_MID.format(
      helo = config["HELO"], identifier=identifier),
    inreply = answermid,
    references = references + " " + answermid,
    rcpt = answeraddress,
    cc_hdr=cc_header(cc), additional_rcpt=cc_rcpt(cc),
    identifier = identifier,
    actdesc = act_desc(extra_files=cmd_extra_files, identifier=identifier),
  )

  _, token_triples = generate_tokens(
    cmds=commands, identifier=identifier, count=0,
    validuntil = referencetime + TOKEN_EXTENDED_VALIDITY,
    secret = config["secret"],
  )

  token_units = []
  tokendict = {}
  token_files = {}
  for mail, use, token in token_triples:
    token_files["token-%s" % (simpleQuote(token))] \
      =  "%s\n%s\n" % (use, mail)
  for mail, use, token in token_triples:
    if mail not in tokendict:
      tokendict[mail] = ''
    tokendict[mail] += TOKEN_DESC[use].format(
      unit = "%s/MAIN" % (identifier,),
      token = token)
  tokenmailcount  = 0
  for mail, descs in tokendict.items():
    tokenmailcount += 1
    token_units.append(Unit(
      name = "token%s" % (tokenmailcount,),
      payload = TOKEN_MAIL.format(
        helo = config["HELO"],
        rcpt = mail,
        inreply = mid,
        references = references,
        mid = TOKEN_MAIL_MID.format(identifier=identifier,
                                    name="token%s" % (tokenmailcount,),
                                    helo = config["HELO"]),
        mailfrom = config["FROM"],
        descs = descs,
        actdesc = act_desc(extra_files=cmd_extra_files, identifier=identifier),
      )))

  payload = scheduledmail.util.joinBsmtp(bsmtp, allsentbsmtp)
  final_cmds = [cmd[1:] for cmd in commands if cmd[0] == "finally"]
  if final_cmds:
    finalbsmtp = generate_final_mail(
      identifier = identifier,
      name = "MAIN",
      final_cmds = final_cmds,
      answeraddress = answeraddress,
      config = config,
      references = references,
    )
    payload = scheduledmail.util.joinBsmtp(payload, finalbsmtp)

  return [answer,
          Unit(payload=payload,
               epoch=time,
               extra_files=dict(**cmd_extra_files, **token_files))] \
               + token_units + announceunits

def get_after(*, commands, name, identifier):
  result = []
  for cmd in commands:
    if len(cmd) >= 2 and cmd[0] == "after" and cmd[1] == name:
      result.append([x.replace("%%THIS%%", identifier) for x in cmd[2:]])
  return result

def get_max_offset(commands):
  maxoffset = 0
  for cmd in commands:
    if cmd[0] == "offset":
      offsetstring = cmd[2:]
      if len(offsetstring) == 1:
        try:
          offset = int(offsetstring[0])
          maxoffset = max(maxoffset, offset)
        except ValueError:
          pass
      if len(offsetstring) >= 2 and offsetstring[1] in TIME_UNITS:
        try:
          offset = int(offsetstring[0]) * TIME_UNITS[offsetstring[1]]
          maxoffset = max(maxoffset, offset)
        except ValueError:
          pass
  return maxoffset

def get_offset(*, commands, name):
  warnings = []
  offset = 0
  for cmd in commands:
    if len(cmd) >= 2 and cmd[0] == "offset" and cmd[1] == name:
      if len(cmd) == 2:
        warnings.append("offset for %s: ignoring missing value" % (name,))
        continue
      offsetstring = cmd[2:]
      unit = 1
      if len(offsetstring) >= 2:
        if offsetstring[1] in TIME_UNITS:
          unit = TIME_UNITS[offsetstring[1]]
        else:
          warnings.append("offset for %s: ignoring unkown time unit %s"
                          % (name, offsetstring[1]))
      try:
        offset = int(offsetstring[0]) * unit
      except ValueError:
        warnings.append("offset for %s: cannot parse number %s"
                        % (name, offsetstring[0]))
  return Success(offset, warnings=warnings)

def auto_bcc(commands):
  result = True
  for cmd in commands:
    if cmd[0] == "noautobcc":
      result = False
  return result

def process_attachment(
    *, part, count, config, answeraddress, cc, answermid,
    epoch, mid, identifier, extra_files, commands, references):
  warnings = []
  name = "%d" % (count,)
  if count == 1:
    name = "MAIN"
  # Step down to the actual mail
  part = part.get_payload()[0]

  # determine receipients
  tos = part.get_all("to", [])
  ccs = part.get_all("cc", [])
  bccs = part.get_all("bcc", [])
  rcptswithnames = email.utils.getaddresses(tos + ccs + bccs)
  rcpts = [mail for (_, mail) in rcptswithnames]

  part_desc = []
  part_desc.append("RCPT TO: " + ", ".join(rcpts))
  if "From" in part:
    part_desc.append("From: " + part.get("From"))
  if "Subject" in part:
    part_desc.append("Subject: " + part.get("Subject"))

  if len(rcpts) == 0:
    rcpts = [answeraddress]
    warnings.append("Found mail attachement without recipients")

  # add implict receipients if not present, unless requested not to do so
  if auto_bcc(commands):
    for stdrcpt in cc + [answeraddress]:
      if stdrcpt not in rcpts:
        rcpts.append(stdrcpt)

  offsetsucc = get_offset(commands=commands, name=name)
  warnings.extend(offsetsucc.warnings)

  # purge headers
  for header in PURGE_HEADERS:
    del part[header]

  if not part.get("Message-ID"):
    part["Message-ID"] = PAYLOAD_MID.format(
      helo = config["HELO"],
      identifier = identifier,
       name = name,
    )

  msg = part.as_string()
  payload_bsmtp_lines = [
    "HELO %s" % (config["HELO"],),
    "MAIL FROM: %s" % (config["FROM"],),
  ]
  for rcpt in rcpts:
    payload_bsmtp_lines.append("RCPT TO: %s" % (rcpt,))
  payload_bsmtp_lines.append("DATA")

  for line in msg.splitlines():
    if line.startswith('.'):
      payload_bsmtp_lines.append(".%s" % (line,))
    else:
      payload_bsmtp_lines.append(line)
  payload_bsmtp_lines.append(".")
  payload_bsmtp_lines.append("QUIT")

  payload_bsmtp = "\n".join(payload_bsmtp_lines) + "\n"

  donemid = SCHEDULED_MAIL_PART_SENT_MID.format(
    identifier=identifier,
    name=name,
    helo=config["HELO"])

  donebsmtp = SCHEDULED_MAIL_SENT.format(
    helo = config["HELO"],
    rcpt = answeraddress,
    cc_hdr=cc_header(cc), additional_rcpt=cc_rcpt(cc),
    mailfrom = config["FROM"],
    mid = donemid,
    inreply = answermid,
    references = references + " " + answermid,
    identifier = "%s/%s" % (identifier, name),
    actdesc = act_desc(extra_files=extra_files, identifier=identifier),
  )

  payload = scheduledmail.util.joinBsmtp(
    payload_bsmtp, donebsmtp)

  final_cmds = get_after(commands=commands, name=name, identifier=identifier)
  if final_cmds:
    finalbsmtp = generate_final_mail(
      identifier = identifier,
      name = name,
      final_cmds = final_cmds,
      answeraddress = answeraddress,
      config = config,
      references = references,
    )
    payload = scheduledmail.util.joinBsmtp(
      payload, finalbsmtp)

  return Unit(payload, epoch=epoch + offsetsucc.payload, name=name,
              extra_files=extra_files, desc=part_desc), warnings


def process_internal(*, mail, mid, identifier, config, references):
  """
  Process the file for a given identifier (that may or may not be
  needed to generate answer mails) and return a list of units.
  """
  answeraddress = email.utils.parseaddr(mail.get("From"))[1]
  answermid = SUBMISSION_ANSWER_MID.format(
    identifier=identifier, helo = config["HELO"])

  if not mail.is_multipart():
    return [failUnit(
      rcpt = answeraddress, cc=[],
      inreply = mid, references=references, config = config, mid = answermid,
      messages = ["Nothing found to schedule. Please provide the payload as"
                  + " attached mails or as a single .bsmtp file"],
    )]
  bsmtp = None
  filenames = []
  for part in mail.get_payload():
    name = part.get_filename()
    if name is None:
      continue
    if name.split('.')[1] == "bsmtp":
      filenames.append(name)
      bsmtp = part.get_content()

  if len(filenames) > 1:
    return (mid,[failUnit(
      rcpt = answeraddress, cc=cc,
      inreply = mid, references=references, config=config, mid = answermid,
      messages =["Expecting at most one .bsmtp attachment; found %s"
                 % (filenames,)],
    )])
  if len(filenames) == 1:
    return process_bsmtp(identifier=identifier, config=config, bsmtp=bsmtp,
                         answeraddress=answeraddress, answermid=answermid,
                         mail=mail, mid=mid, references=references)

  body = mail.get_body().get_content()
  commands_success = scheduledmail.parser.parse_command_block(
    body, commands=["schedule", "me", "cc", "suspend", "description", "token",
                    "announce", "announcemsg", "wantlgtm", "after", "offset",
                    "noautobcc", "tokenmain",
    ],
    alternatives=CMD_ALTERNATIVES)
  commands = commands_success.payload
  warnings = commands_success.warnings
  time_success = get_time(commands)
  warnings.extend(time_success.warnings)
  time = time_success.payload
  me_success = get_me(commands)
  warnings.extend(me_success.warnings)
  if me_success.payload:
    answeraddress = me_success.payload
  cc = get_cc(commands)

  if time is None:
    return [failUnit(
      rcpt = answeraddress, cc=cc,
      inreply = mid, references=references, config = config, mid = answermid,
      messages = ["Could not determine the time to schedule the request"]
                  + warnings,
    )]

  referencetime, timewarnings = not_before_now(time)
  warnings.extend(timewarnings)

  units = []
  names = []
  unit_count = 0
  token_count = 0

  maxoffset = get_max_offset(commands)
  tokendict = {}
  # A map from email to string with the string telling the user about
  # the tokens granted to them.
  cmd_extra_files = extra_files(commands)
  for part in mail.get_payload():
    if part.get_content_type() == "message/rfc822":
      unit_count += 1
      token_count, token_triples = generate_tokens(
        cmds=commands, identifier=identifier, count=token_count,
        validuntil = referencetime + maxoffset + TOKEN_EXTENDED_VALIDITY,
        secret = config["secret"],
        ismain = unit_count == 1,
      )
      token_files = {}
      for mail, use, token in token_triples:
        token_files["token-%s" % (simpleQuote(token))] \
          =  "%s\n%s\n" % (use, mail)
      unit, unitwarnings = process_attachment(
        part=part, count=unit_count, config=config, answeraddress=answeraddress,
        answermid=answermid, cc=cc, epoch=time, mid=mid, references=references,
        extra_files=dict(**cmd_extra_files, **token_files),
        identifier=identifier, commands=commands)
      units.append(unit)
      warnings.extend(unitwarnings)
      names.append(unit.name)
      for mail, use, token in token_triples:
        if mail not in tokendict:
          tokendict[mail] = ''
        tokendict[mail] += TOKEN_DESC[use].format(
          unit = "%s/%s" % (identifier, unit.name),
          token = token)
      if unit.desc:
        unit_explain = PART_DESC_ANNOUNCE.format(
          unit = "%s/%s" % (identifier, unit.name))
        unit_explain += "\n".join([" | " + item for item in unit.desc])
        unit_explain += "\n\n\n"
        for mail in set([mail for mail, use, token in token_triples]):
          tokendict[mail] += unit_explain

  if unit_count == 0:
    return [failUnit(
      rcpt = answeraddress, cc=cc,
      inreply = mid, references=references, config=config, mid = answermid,
      messages = ["Neither message/rfc822 nor bsmtp attachment found"])]

  announceunits_succ = generateAnnounce(epoch=time, identifier=identifier,
                                        cmds=commands, config=config,
                                        answeraddress=answeraddress,
                                        references=references)
  warnings.extend(announceunits_succ.warnings)
  announceunits = announceunits_succ.payload

  names.extend([au.name for au in announceunits])

  answer = answerUnit(
    rcpt=answeraddress, cc=cc, inreply=mid, references=references,
    config=config, mid=answermid,
    identifier=identifier, time=time, names=names,
    extrafiles=cmd_extra_files,
    warnings=warnings)

  token_units = []
  tokenmailcount = 0
  for mail, descs in tokendict.items():
    tokenmailcount += 1
    token_units.append(Unit(
      name = "token%s" % (tokenmailcount,),
      payload = TOKEN_MAIL.format(
        helo = config["HELO"],
        rcpt = mail,
        inreply = mid,
        references = references,
        mid = TOKEN_MAIL_MID.format(identifier=identifier,
                                    name="token%s" % (tokenmailcount,),
                                    helo = config["HELO"]),
        mailfrom = config["FROM"],
        descs = descs,
        actdesc = act_desc(extra_files=cmd_extra_files, identifier=identifier),
      )))


  return ([answer] + units + token_units + announceunits)


def main(argv):
  # In this service, we activate payload and answer in one act
  # so we are atomic. In particular retry does not harm, despite resubmit
  return general_service.std_service_main(
    process_internal=process_internal, argv=argv,
    exit_could_not_answer=EXIT_TEMP_FAIL)

if __name__ == "__main__":
  sys.exit(main(sys.argv))
