#!/usr/bin/env bats

load '../../bats-support/load'
load '../../bats-file/load'

function setup {
    TEST_TEMP_DIR="$(temp_make)"
    BATSLIB_FILE_PATH_REM="#${TEST_TEMP_DIR}"
    BATSLIB_FILE_PATH_ADD='<temp>'

    TEST_BIN="$(temp_make)"
    cat >$TEST_BIN/rsmtp <<EOF
#!/bin/sh
cat >> $TEST_BIN/log
EOF
    chmod +x $TEST_BIN/rsmtp
}

function teardown {
    temp_del "$TEST_TEMP_DIR"
    temp_del "$TEST_BIN"
}


@test "Invoking without argument prints an error" {
    run ./mail-schedule-do.pl

    echo $status
    [ "$status" -ne 0 ]
    echo $output
    [ "$output" = "Usage ./mail-schedule-do.pl <scheduledir>" ]
}

@test "Invoking on an empty dir is succesfull" {
    run ./mail-schedule-do.pl $TEST_TEMP_DIR
    [ "$status" -eq 0 ]
}

@test "Exit nonzero, if scheduledir does not exist" {
    # We want to use our 'rsmtp'
    PATH=$TEST_BIN:$PATH
    export PATH

    run ./mail-schedule-do.pl ${TEST_TEMP_DIR}_does_not_exist
    [ "$status" -ne 0 ]
}

@test "Send mail if time is older than now" {
    # We want to use our 'rsmtp'
    PATH=$TEST_BIN:$PATH
    export PATH

    local ACT=`date +%s`.$$
    mkdir $TEST_TEMP_DIR/$ACT
    local UNIT=`date -d '-1 second' +%s`.MAIN
    mkdir $TEST_TEMP_DIR/$ACT/$UNIT
    touch $TEST_TEMP_DIR/$ACT/$UNIT/payload.bsmtp

    run ./mail-schedule-do.pl $TEST_TEMP_DIR
    [ "$status" -eq 0 ]
    assert_file_not_exist $TEST_TEMP_DIR/$ACT
    run sh -c "ls '$TEST_TEMP_DIR' | grep -q '^T-'"
    [ "$status" -eq 1 ]

    rm -rf $TEST_TEMP_DIR/$ACT
}

@test "Don't delete act if there is still a unit pending" {
    # We want to use our 'rsmtp'
    PATH=$TEST_BIN:$PATH
    export PATH

    local ACT=`date +%s`.$$
    mkdir $TEST_TEMP_DIR/$ACT
    local UNIT=`date -d '-1 second' +%s`.early
    mkdir $TEST_TEMP_DIR/$ACT/$UNIT
    touch $TEST_TEMP_DIR/$ACT/$UNIT/payload.bsmtp
    local UNIT2=`date -d '+1 day' +%s`.late
    mkdir $TEST_TEMP_DIR/$ACT/$UNIT2
    touch $TEST_TEMP_DIR/$ACT/$UNIT2/payload.bsmtp

    ls -l $TEST_TEMP_DIR/$ACT >&3
    run ./mail-schedule-do.pl $TEST_TEMP_DIR
    [ "$status" -eq 0 ]
    ls -l $TEST_TEMP_DIR/$ACT >&3
    assert_file_exist $TEST_TEMP_DIR/$ACT/
    assert_file_not_exist $TEST_TEMP_DIR/$ACT/$UNIT
    assert_file_exist $TEST_TEMP_DIR/$ACT/$UNIT2


    rm -rf $TEST_TEMP_DIR/$ACT
}

@test "Delete act if there is no unit in it" {
    # We want to use our 'rsmtp'
    PATH=$TEST_BIN:$PATH
    export PATH

    local ACT=`date +%s`.$$
    mkdir $TEST_TEMP_DIR/$ACT

    run ./mail-schedule-do.pl $TEST_TEMP_DIR
    [ "$status" -eq 0 ]

    assert_file_not_exist $TEST_TEMP_DIR/$ACT

    rm -rf $TEST_TEMP_DIR/$ACT
}

@test "Don't delete a temporary act, when the owning process is still alive" {
    # We want to use our 'rsmtp'
    PATH=$TEST_BIN:$PATH
    export PATH

    local ACT=T-`date +%s`.$$
    mkdir $TEST_TEMP_DIR/$ACT

    run ./mail-schedule-do.pl $TEST_TEMP_DIR
    [ "$status" -eq 0 ]

    assert_file_exist $TEST_TEMP_DIR/$ACT

    rm -rf $TEST_TEMP_DIR/$ACT
}

@test "Delete a temporary act, when the owning process is dead" {
    # We want to use our 'rsmtp'
    PATH=$TEST_BIN:$PATH
    export PATH

    local ACT=T-`date +%s`.$((`cat /proc/sys/kernel/pid_max` + 1))
    mkdir $TEST_TEMP_DIR/$ACT

    run ./mail-schedule-do.pl $TEST_TEMP_DIR
    [ "$status" -eq 0 ]

    assert_file_not_exist $TEST_TEMP_DIR/$ACT

    rm -rf $TEST_TEMP_DIR/$ACT
}

@test "Don't delete a temporary unit, when the owning process is still alive" {
    # We want to use our 'rsmtp'
    PATH=$TEST_BIN:$PATH
    export PATH

    local ACT=`date +%s`.$$
    mkdir $TEST_TEMP_DIR/$ACT
    local UNIT=T-`date +%s`.$$
    mkdir $TEST_TEMP_DIR/$ACT/$UNIT

    run ./mail-schedule-do.pl $TEST_TEMP_DIR
    [ "$status" -eq 0 ]

    assert_file_exist $TEST_TEMP_DIR/$ACT/$UNIT

    rm -rf $TEST_TEMP_DIR/$ACT/$UNIT
}

@test "Delete a temporary unit, when the owning process is dead" {
    # We want to use our 'rsmtp'
    PATH=$TEST_BIN:$PATH
    export PATH

    local ACT=`date +%s`.$$
    mkdir $TEST_TEMP_DIR/$ACT
    local UNIT=T-`date +%s`.$((`cat /proc/sys/kernel/pid_max` + 1))
    mkdir $TEST_TEMP_DIR/$ACT/$UNIT

    run ./mail-schedule-do.pl $TEST_TEMP_DIR
    [ "$status" -eq 0 ]

    assert_file_not_exist $TEST_TEMP_DIR/$ACT/$UNIT
}

@test "Don't rename a running unit, when the owning process is still alive" {
    # We want to use our 'rsmtp'
    PATH=$TEST_BIN:$PATH
    export PATH

    local ACT=`date +%s`.$$
    mkdir $TEST_TEMP_DIR/$ACT
    local UNIT=R-`date +%s`.$$-`date +%s`.MAIN
    mkdir $TEST_TEMP_DIR/$ACT/$UNIT

    run ./mail-schedule-do.pl $TEST_TEMP_DIR
    echo $output
    [ "$status" -eq 0 ]

    assert_file_exist $TEST_TEMP_DIR/$ACT/$UNIT
}

@test "Rename a running unit back, when the owning process is dead" {
    # We want to use our 'rsmtp'
    PATH=$TEST_BIN:$PATH
    export PATH

    local ACT=`date +%s`.$$
    mkdir $TEST_TEMP_DIR/$ACT
    local REAL_UNIT=`date +%s`.MAIN
    local UNIT=R-`date +%s`.$((`cat /proc/sys/kernel/pid_max` + 1))-$REAL_UNIT
    mkdir $TEST_TEMP_DIR/$ACT/$UNIT

    run ./mail-schedule-do.pl $TEST_TEMP_DIR
    echo $output
    [ "$status" -eq 0 ]

    assert_file_not_exist $TEST_TEMP_DIR/$ACT/$UNIT
    assert_file_exist $TEST_TEMP_DIR/$ACT/$REAL_UNIT
}

@test "If the rsmtp call fails, the act should stay there" {
    # We want to use our 'rsmtp'
    PATH=$TEST_BIN:$PATH
    export PATH

    local ACT=`date +%s`.$$
    mkdir $TEST_TEMP_DIR/$ACT
    local UNIT=`date -d '-1 second' +%s`.early
    mkdir $TEST_TEMP_DIR/$ACT/$UNIT
    touch $TEST_TEMP_DIR/$ACT/$UNIT/payload.bsmtp
    
    rm $TEST_BIN/rsmtp
    ln -s /bin/false $TEST_BIN/rsmtp

    run ./mail-schedule-do.pl $TEST_TEMP_DIR
    [ "$status" -eq 0 ]

    assert_file_exist $TEST_TEMP_DIR/$ACT/$UNIT
    assert_file_exist $TEST_TEMP_DIR/$ACT/$UNIT/payload.bsmtp

}

@test "Test if rsmtp gets the right payload" {
    # We want to use our 'rsmtp'
    PATH=$TEST_BIN:$PATH
    export PATH

    local ACT=`date +%s`.$$
    mkdir $TEST_TEMP_DIR/$ACT
    local UNIT=`date -d '-1 second' +%s`.early
    mkdir $TEST_TEMP_DIR/$ACT/$UNIT
    echo "Test" >  $TEST_TEMP_DIR/$ACT/$UNIT/payload.bsmtp
    
    run ./mail-schedule-do.pl $TEST_TEMP_DIR
    [ "$status" -eq 0 ]

    [ "`cat $TEST_BIN/log`" = "Test" ]
}

@test "Do not send an unit, when a WANTS_LGTM file exists" {
    # We want to use our 'rsmtp'
    PATH=$TEST_BIN:$PATH
    export PATH

    local ACT=`date +%s`.$$
    mkdir $TEST_TEMP_DIR/$ACT
    local UNIT=`date -d '-1 second' +%s`.early
    mkdir $TEST_TEMP_DIR/$ACT/$UNIT
    echo "Test" >  $TEST_TEMP_DIR/$ACT/$UNIT/payload.bsmtp
    touch $TEST_TEMP_DIR/$ACT/$UNIT/WANT_LGTM

    run ./mail-schedule-do.pl $TEST_TEMP_DIR
    [ "$status" -eq 0 ]

    assert_file_exist $TEST_TEMP_DIR/$ACT/$UNIT
    assert_file_not_exist $TEST_BIN/log
}

@test "Do not send an unit, when a SUSPENDED file exists" {
    # We want to use our 'rsmtp'
    PATH=$TEST_BIN:$PATH
    export PATH

    local ACT=`date +%s`.$$
    mkdir $TEST_TEMP_DIR/$ACT
    local UNIT=`date -d '-1 second' +%s`.early
    mkdir $TEST_TEMP_DIR/$ACT/$UNIT
    echo "Test" >  $TEST_TEMP_DIR/$ACT/$UNIT/payload.bsmtp
    touch $TEST_TEMP_DIR/$ACT/$UNIT/SUSPENDED

    run ./mail-schedule-do.pl $TEST_TEMP_DIR
    [ "$status" -eq 0 ]

    assert_file_exist $TEST_TEMP_DIR/$ACT/$UNIT
    assert_file_not_exist $TEST_BIN/log
}

@test "Don't delete a done/messageid file if the messageid is recent" {
    # We want to use our 'rsmtp'
    PATH=$TEST_BIN:$PATH
    export PATH

    mkdir $TEST_TEMP_DIR/done
    local MESSAGEID=20190530184537_46_GA18786_64_example_46_org
    touch -d "`date -d '-8 days +10 minutes'`" $TEST_TEMP_DIR/done/$MESSAGEID

    run ./mail-schedule-do.pl $TEST_TEMP_DIR

    assert_file_exist $TEST_TEMP_DIR/done/$MESSAGEID
    assert_file_not_exist $TEST_BIN/log
}

@test "Delete a done/messageid file if the messageid is not recent" {
    # We want to use our 'rsmtp'
    PATH=$TEST_BIN:$PATH
    export PATH

    mkdir $TEST_TEMP_DIR/done
    local MESSAGEID=20190530184537_46_GA18786_64_example_46_org
    touch -d "`date -d '-8 days -10 minutes'`" $TEST_TEMP_DIR/done/$MESSAGEID

    run ./mail-schedule-do.pl $TEST_TEMP_DIR
    [ "$status" -eq 0 ]

    assert_file_not_exist $TEST_TEMP_DIR/done/$MESSAGEID
    assert_file_not_exist $TEST_BIN/log
}

@test "Valid tokens should not be deleted" {
    # We want to use our 'rsmtp'
    PATH=$TEST_BIN:$PATH
    export PATH

    mkdir $TEST_TEMP_DIR/muted
    local token_end="_58_9338f7c70e60_58_abb39dba77d0"
    local hex=$(printf '%x' "$((`date -d '+10 minutes' +%s` >> 16))")
    local tokenfile=${hex}${token_end}
    touch $TEST_TEMP_DIR/muted/$tokenfile

    run ./mail-schedule-do.pl $TEST_TEMP_DIR
    [ "$status" -eq 0 ]

    assert_file_exist $TEST_TEMP_DIR/muted/$tokenfile
    assert_file_not_exist $TEST_BIN/log
}

@test "Invalid tokens should be deleted" {
    # We want to use our 'rsmtp'
    PATH=$TEST_BIN:$PATH
    export PATH

    mkdir $TEST_TEMP_DIR/muted
    local token_end="_58_9338f7c70e60_58_abb39dba77d0"
    local hex=$(printf '%x' "$((`date -d '-1 day' +%s` >> 16))")
    local tokenfile=${hex}${token_end}
    touch $TEST_TEMP_DIR/muted/$tokenfile

    run ./mail-schedule-do.pl $TEST_TEMP_DIR
    [ "$status" -eq 0 ]

    assert_file_not_exist $TEST_TEMP_DIR/muted/$tokenfile
    assert_file_not_exist $TEST_BIN/log
}
