import unittest
import tempfile

import unittest.mock
from unittest.mock import patch

import os

import scheduledmail.lgtm
import scheduledmail.mailtoken
import scheduledmail.parser
import scheduledmail.schedule
import scheduledmail.simple_submit
import scheduledmail.submit
import scheduledmail.util

from scheduledmail.unit import Unit
from scheduledmail.common import Success, ScheduleError

class TestUtil(unittest.TestCase):
  def testJoinBsmtp(self):
    self.assertEqual(
     scheduledmail.util.joinBsmtp(
       "HELO foo.example.com\nMAIL FROM: foo\n...\nfoo\n.\nQUIT\n",
       "HELO bar.example.com\nMAIL FROM: bar\n...\nbar\n.\nQUIT\n"),
       "HELO foo.example.com\nMAIL FROM: foo\n...\nfoo\n.\nMAIL FROM: bar\n...\nbar\n.\nQUIT\n"),
  def testSimpleQuote(self):
    self.assertEqual(
      scheduledmail.util.simpleQuote(
        "<20190530152358.GA24618@mymachine.example.org>"),
      "_60_20190530152358_46_GA24618_64_mymachine_46_example_46_org_62_")

class TestSuccess(unittest.TestCase):
  def testPrintName(self):
    successA = Success("A", warnings = ["warn"], dirruns = ["somedir"])
    self.assertEqual("%s" % successA, "A")

  def testThenAfter(self):
    successA = Success("A", warnings = ["warn A"], dirruns = ["somedir A"])
    successB = Success("B", warnings = ["warn B"], dirruns = ["somedir B"])
    aThenB = successA.then(successB)
    self.assertEqual(aThenB.payload, "B")
    self.assertEqual(aThenB.warnings, ["warn A", "warn B"])
    self.assertEqual(aThenB.dirruns, ["somedir A", "somedir B"])
    aAfterB = successA.after(successB)
    self.assertEqual(aAfterB.payload, "A")
    self.assertEqual(aAfterB.warnings, ["warn B", "warn A"])
    self.assertEqual(aAfterB.dirruns, ["somedir B", "somedir A"])

  def testForM_(self):
    f = lambda x : Success(x, warnings = [x])
    xs = ["A", "B", "C"]
    forM = Success.forM_(xs, f)
    self.assertEqual(forM.payload, None)
    self.assertEqual(forM.warnings, ["A", "B", "C"])


class TestToken(unittest.TestCase):
  def test_time_prefix(self):
    t = 1559411750
    day = 24 * 60 * 60
    prefix = scheduledmail.mailtoken.timeprefix(t)
    self.assertTrue(scheduledmail.mailtoken.timeprefix_valid(
      prefix, now=t - day))
    self.assertTrue(scheduledmail.mailtoken.timeprefix_valid(prefix, now=t))
    self.assertFalse(scheduledmail.mailtoken.timeprefix_valid(
      prefix, now=t + day))
    # should not crash on malformed input
    self.assertFalse(scheduledmail.mailtoken.timeprefix_valid(
      " ^!$asfd", now=t))

  def test_token_valid(self):
    validuntil = 1560117600
    day = 24 * 60 * 60
    identifier = "1559418304.4275"
    secret = "ServerSecret!"
    mytoken = scheduledmail.mailtoken.generate_token(
      secret=secret, identifier=identifier, count=1, validuntil=validuntil)
    othertoken = scheduledmail.mailtoken.generate_token(
      secret="badSecret", identifier=identifier, count=1, validuntil=validuntil)
    self.assertTrue(scheduledmail.mailtoken.token_valid(
      token=mytoken, secret=secret, now=validuntil - 10 * day))
    self.assertTrue(scheduledmail.mailtoken.token_valid(
      token=mytoken, secret=secret, now=validuntil - day))
    self.assertTrue(scheduledmail.mailtoken.token_valid(
      token=mytoken, secret=secret, now=validuntil))
    self.assertFalse(scheduledmail.mailtoken.token_valid(
      token=mytoken, secret=secret, now=validuntil + day))
    self.assertFalse(scheduledmail.mailtoken.token_valid(
      token=othertoken, secret=secret, now=validuntil))
    self.assertFalse(scheduledmail.mailtoken.token_valid(
      token=othertoken, secret=secret, now=validuntil))
    # should not crash on malformed input
    self.assertFalse(scheduledmail.mailtoken.token_valid(
      token="5f72::::", secret=secret, now=validuntil))
    self.assertFalse(scheduledmail.mailtoken.token_valid(
      token="5f72::", secret=secret, now=validuntil))
    self.assertFalse(scheduledmail.mailtoken.token_valid(
      token="5f72:", secret=secret, now=validuntil))
    self.assertFalse(scheduledmail.mailtoken.token_valid(
      token="5f72", secret=secret, now=validuntil))
    self.assertFalse(scheduledmail.mailtoken.token_valid(
      token="5f72:/asdf=:", secret=secret, now=validuntil))

class TestLGTM(unittest.TestCase):
  def test_normal_decrement(self):
    self.assertEqual(scheduledmail.lgtm.do_lgtm(want_lgtm="1"),
                     None)
    self.assertEqual(scheduledmail.lgtm.do_lgtm(want_lgtm="whatever"),
                     None)
    self.assertEqual(scheduledmail.lgtm.do_lgtm(want_lgtm=""),
                     None)
    self.assertEqual(scheduledmail.lgtm.do_lgtm(want_lgtm="2"),
                     "1")
    self.assertEqual(scheduledmail.lgtm.do_lgtm(want_lgtm="2\n"),
                     "1")
    self.assertEqual(scheduledmail.lgtm.do_lgtm(want_lgtm="3"),
                     "2")
    self.assertEqual(scheduledmail.lgtm.do_lgtm(want_lgtm=None),
                     None)
  def test_normal_count(self):
    self.assertEqual(scheduledmail.lgtm.do_lgtm(want_lgtm=None, count="-1"),
                     "1")
    self.assertEqual(scheduledmail.lgtm.do_lgtm(want_lgtm="yes please",
                                                count="-1"),
                     "2")
    self.assertEqual(scheduledmail.lgtm.do_lgtm(want_lgtm="3", count="100"),
                     None)
    self.assertEqual(scheduledmail.lgtm.do_lgtm(want_lgtm="3", count="2"),
                     "1")
    self.assertEqual(scheduledmail.lgtm.do_lgtm(want_lgtm="3", count="thanks"),
                     "2")

  def test_format_lgtm_status(self):
    self.assertEqual(scheduledmail.lgtm.format_lgtm_status(None),
                     "No more reviews required")
    self.assertEqual(scheduledmail.lgtm.format_lgtm_status(""),
                     "1 review required")
    self.assertEqual(scheduledmail.lgtm.format_lgtm_status("1"),
                     "1 review required")
    self.assertEqual(scheduledmail.lgtm.format_lgtm_status("1\n"),
                     "1 review required")
    self.assertEqual(scheduledmail.lgtm.format_lgtm_status("  1  \n"),
                     "1 review required")
    self.assertEqual(scheduledmail.lgtm.format_lgtm_status("I want review"),
                     "1 review required")
    self.assertEqual(scheduledmail.lgtm.format_lgtm_status("2"),
                     "2 reviews required")

class TestParser(unittest.TestCase):
  def testSimple(self):
    parse = scheduledmail.parser.parse_command_block(
      "schedule 2019-06-01 12:00\nthanks\n\nWe should really answer...\n",
      commands = ["schedule"])
    self.assertEqual(parse.payload, [['schedule', '2019-06-01', '12:00']])
    self.assertEqual(parse.warnings, [])

  def testMultiple(self):
    parse = scheduledmail.parser.parse_command_block(
      "cmd A\ncmd B\nthanks\ncmd C", commands = ["cmd"])
    self.assertEqual(parse.payload, [['cmd', 'A'], ['cmd', 'B']])
    self.assertEqual(parse.warnings, [])

  def testWarnings(self):
    parse = scheduledmail.parser.parse_command_block(
      "cmd A\nbad\ncmd B\nthanks\ncmd C",
      commands = ["cmd"])
    self.assertEqual(parse.payload, [['cmd', 'A'], ['cmd', 'B']])
    self.assertTrue(len(parse.warnings) > 0)

  def testBailOut(self):
    parse = scheduledmail.parser.parse_command_block(
      """
cmd A

unrelated
unrelated
unrelated
unrelated
unrelated

Too late to add a
cmd X
      """, commands = ["cmd"])
    self.assertEqual(parse.payload, [['cmd', 'A']])
    self.assertTrue(len(parse.warnings) > 0)

  def testUpperCase(self):
    parse = scheduledmail.parser.parse_command_block(
      "CMD foo BAR\nTHANKS\ncmd ignoreme\n",
      commands=["cmd"])
    self.assertEqual(parse.payload, [['cmd', 'foo', 'BAR']])
    self.assertEqual(parse.warnings, [])

  def testComment(self):
    parse = scheduledmail.parser.parse_command_block(
      """
>fds
> thanks
cmd A
# This is a comment
cmd B
thanks
""", commands=["cmd"])
    self.assertEqual(parse.warnings, [])
    self.assertEqual(parse.payload, [['cmd', 'A'], ['cmd', 'B']])

  def testAlternatives(self):
    parse = scheduledmail.parser.parse_command_block(
      """
lol a b
thanks
""", commands=["cmd"], alternatives={'lol': 'cmd'})
    self.assertEqual(parse.warnings, [])
    self.assertEqual(parse.payload, [['cmd', 'a', 'b']])

  def testAlternativesWithoutCmd(self):
    parse = scheduledmail.parser.parse_command_block(
      """
lol a b
thanks
""", commands=['test'], alternatives={'lol': 'cmd'})
    self.assertEqual(parse.payload, [])
    self.assertTrue(len(parse.warnings) > 0)

class TestExtraFiles(unittest.TestCase):
  def test_multi(self):
    self.assertEqual(
      scheduledmail.submit.extra_files([
        ['cc', 'you@example.com'],
        ['description', 'This', 'is', 'fun'],
        ['suspend'],
        ['description'],
        ['cc', 'me@example.org'],
        ['description', 'and', 'more', 'fun'],
      ]),
      { 'CC': 'you@example.com\nme@example.org\n',
        'DESCRIPTION' : 'This is fun\n\nand more fun\n',
        'SUSPENDED' : '\n',
      },
    )

class TestUnquoteAndSplit(unittest.TestCase):
  def test_unquote_and_split(self):
    self.assertEqual(
      scheduledmail.simple_submit.unquote_and_split("foo%40example.com"),
      ["foo@example.com"])
    self.assertEqual(
      scheduledmail.simple_submit.unquote_and_split("foo%40example:com"),
      ["foo@example.com"])
    self.assertEqual(
      scheduledmail.simple_submit.unquote_and_split("foo:bar:baz%40example:com"),
      ["foo.bar.baz@example.com"])
    self.assertEqual(
      scheduledmail.simple_submit.unquote_and_split("foo%40example%3acom"),
      ["foo@example:com"])
    self.assertEqual(
      scheduledmail.simple_submit.unquote_and_split(
        "foo%40example.com%00bar%40example.com"),
      ["foo@example.com", "bar@example.com"])
  def test_time_rcpt_split(self):
    self.assertEqual(
      scheduledmail.simple_submit.split_time_rcpt("05241800"),
      ("05241800", []))
    self.assertEqual(
      scheduledmail.simple_submit.split_time_rcpt("05241800-foo%40example.com"),
      ("05241800", ["foo@example.com"]))
    self.assertEqual(
      scheduledmail.simple_submit.split_time_rcpt(
        "05241800-foo-bar-baz%40example.com"),
      ("05241800", ["foo-bar-baz@example.com"]))
    self.assertEqual(
      scheduledmail.simple_submit.split_time_rcpt(
        "05241800-foo%40example.com%00bar%40example.org"),
      ("05241800", ["foo@example.com", "bar@example.org"]))


class TestSchedule(unittest.TestCase):

  def setUp(self):
    self.scheduledir = tempfile.TemporaryDirectory()
    self.schedule_config = {'scheduledir': self.scheduledir.name,
                            'at': '/bin/true'}


  def tearDown(self):
    self.scheduledir.cleanup()


  def test_new_entry_simple(self):
    test_unit = Unit("test_payload\n")
    new_entry = scheduledmail.schedule.new_entry("A", units=[test_unit],
                                                 config=self.schedule_config)
    self.assertEqual(new_entry, Success("A"))

  @patch.object(os.path, 'isdir')
  def test_error_scheduledir_does_not_exist(self, is_dir):
    is_dir.return_value = False

    test_unit = Unit("test_payload\n")

    with self.assertRaisesRegex(ScheduleError, 'does not exist'):
      new_entry = scheduledmail.schedule.new_entry("A", units=[test_unit],
                                                   config=self.schedule_config)

  @patch.object(os, 'mkdir')
  def test_error_if_failed_to_create_act(self, mkdir):
    mkdir.side_effect = Exception('TEST')

    test_unit = Unit("test_payload\n")

    with self.assertRaisesRegex(ScheduleError,
                                'Error creating schedule on disk'):
      new_entry = scheduledmail.schedule.new_entry("A", units=[test_unit],
                                                   config=self.schedule_config)

  @patch.object(os, 'mkdir')
  def test_empty_units_do_nothing(self, mkdir):
    new_entry = scheduledmail.schedule.new_entry("A", units=[],
                                                 config=self.schedule_config)
    self.assertEqual(new_entry, Success("A"))
    mkdir.assert_not_called()

  @patch.object(os, 'mkdir')
  def test_none_units_do_nothing(self, mkdir):
    new_entry = scheduledmail.schedule.new_entry("A",
                                                 config=self.schedule_config)
    self.assertEqual(new_entry, Success("A"))
    mkdir.assert_not_called()


if __name__ == '__main__':
    unittest.main()
